# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import logging
import os
import sys
import time

import pytest
import socket
from subprocess import Popen, PIPE

from tango import DevState, DbData, DbDatum
from tango.utils import ensure_binary
from tango.test_utils import wait_for_proxy

# Helpers

MAX_STARTUP_TIME_SEC = 30.0


def start_database(port, inst):
    python = sys.executable
    tests_directory = os.path.dirname(__file__)
    cmd = (
        f"{python} -u -m tango.databaseds.database"
        f" --host=127.0.0.1 --port={port}"
        f" --logging_level=2 {inst}"
    )
    env = os.environ.copy()
    env["PYTANGO_DATABASE_NAME"] = ":memory:"  # Don't write to disk
    logging.debug("Starting databaseds subprocess...")
    proc = Popen(
        cmd.split(), cwd=tests_directory, stdout=PIPE, bufsize=1, text=True, env=env
    )
    logging.debug("Waiting for databaseds to startup...")
    try:
        wait_for_tango_server_startup(proc)
    except RuntimeError:
        proc.terminate()
        raise

    return proc


def wait_for_tango_server_startup(proc):
    t0 = time.time()
    for line in proc.stdout:
        now = time.time()
        elapsed = now - t0
        print(line, end="")
        if "Ready to accept request" in line:
            logging.debug(f"Databaseds startup complete after {elapsed:.1f} sec")
            break
        if elapsed > MAX_STARTUP_TIME_SEC:
            msg = f"Databaseds startup timed out after {elapsed:.1f} sec"
            logging.error(msg)
            raise RuntimeError(msg)


def get_open_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


@pytest.fixture()
def tango_database_test():
    port = get_open_port()
    inst = 2

    proc = start_database(port, inst)
    logging.debug("Waiting for databaseds proxy...")
    proxy = wait_for_proxy(f"tango://127.0.0.1:{port}/sys/database/2")
    logging.debug("Databaseds proxy is ready")

    yield proxy

    proc.terminate()
    logging.debug("Terminated databaseds")
    print("Remaining databaseds output:")
    for line in proc.stdout:
        print(line, end="")


# Tests
def test_ping(tango_database_test):
    duration = tango_database_test.ping(wait=True)
    assert isinstance(duration, int)


def test_status(tango_database_test):
    assert tango_database_test.status() == "The device is in ON state."


def test_state(tango_database_test):
    assert tango_database_test.state() == DevState.ON


def test_device_property(tango_database_test):
    def get_name_value(num):
        val = f"value {num}"
        return f"property {num}", val, ensure_binary(val, encoding="latin-1")

    ## put_property overloads:
    # With DbDatum as argument:
    property_name, _, property_value_encoded = get_name_value(1)
    ddatum = DbDatum(property_name)
    ddatum.value_string.append(property_value_encoded)
    tango_database_test.put_property(ddatum)
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 1
    assert return_property_list[0] == property_name

    # With DbData as argument:
    property_name, _, property_value_encoded = get_name_value(2)
    ddatum = DbDatum(property_name)
    ddatum.value_string.append(property_value_encoded)
    ddata = DbData()
    ddata.append(ddatum)
    tango_database_test.put_property(ddata)
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 2
    assert return_property_list[1] == property_name

    # With dict[name: value] as argument:
    property_name, property_value, _ = get_name_value(3)
    tango_database_test.put_property({property_name: property_value})
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 3
    assert return_property_list[2] == property_name

    # With dict[name: DaDatum] as argument:
    property_name, _, property_value_encoded = get_name_value(4)
    ddatum = DbDatum(property_name)
    ddatum.value_string.append(property_value_encoded)
    tango_database_test.put_property({"some_name": ddatum})
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 4
    assert return_property_list[3] == property_name

    # With dict[str: list[str]] as argument:
    property_name, property_value_str, _ = get_name_value(5)
    tango_database_test.put_property({property_name: [property_value_str]})
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 5
    assert return_property_list[4] == property_name

    # With list[DaDatum] as argument:
    property_name_1, _, property_value_encoded_1 = get_name_value(6)
    ddatum_1 = DbDatum(property_name_1)
    ddatum_1.value_string.append(property_value_encoded_1)
    property_name_2, _, property_value_encoded_2 = get_name_value(7)
    ddatum_2 = DbDatum(property_name_2)
    ddatum_2.value_string.append(property_value_encoded_2)
    property_name_3, _, property_value_encoded_3 = get_name_value(8)
    ddatum_3 = DbDatum(property_name_3)
    ddatum_3.value_string.append(property_value_encoded_3)
    property_name_4, _, property_value_encoded_4 = get_name_value(9)
    ddatum_4 = DbDatum(property_name_4)
    ddatum_4.value_string.append(property_value_encoded_4)
    tango_database_test.put_property([ddatum_1, ddatum_2, ddatum_3, ddatum_4])
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 9
    assert return_property_list[5] == property_name_1
    assert return_property_list[6] == property_name_2
    assert return_property_list[7] == property_name_3
    assert return_property_list[8] == property_name_4

    # With empty dict as argument (should not raise):
    tango_database_test.put_property({})

    # With empty list as argument (should not raise):
    tango_database_test.put_property([])

    # With empty tuple as argument (should not raise):
    tango_database_test.put_property(())

    ## get_property overloads:
    return_property = tango_database_test.get_property(property_name)
    assert return_property[property_name][0] == property_value_str

    dbdata = DbData()
    assert len(dbdata) == 0
    return_property = tango_database_test.get_property(property_name, dbdata)
    assert return_property[property_name][0] == property_value_str
    assert len(dbdata) == 1

    datum = DbDatum(property_name)
    return_property = tango_database_test.get_property(datum)
    assert return_property[property_name][0] == property_value_str

    data = DbData()
    data.append(DbDatum(property_name))
    return_property = tango_database_test.get_property(data)
    assert return_property[property_name][0] == property_value_str

    return_property = tango_database_test.get_property([property_name])
    assert return_property[property_name][0] == property_value_str

    dbdata = DbData()
    assert len(dbdata) == 0
    return_property = tango_database_test.get_property([property_name], dbdata)
    assert return_property[property_name][0] == property_value_str
    assert len(dbdata) == 1

    return_property = tango_database_test.get_property([DbDatum(property_name)])
    assert return_property[property_name][0] == property_value_str

    return_property = tango_database_test.get_property({property_name: None})
    assert return_property[property_name][0] == property_value_str

    property_name, _, _ = get_name_value(5)
    ddatum = DbDatum(property_name)
    return_property = tango_database_test.get_property({"some_name": ddatum})
    assert return_property[property_name][0] == property_value_str

    # With empty dict as argument (should give empty dict):
    return_property = tango_database_test.get_property({})
    assert return_property == {}

    # With empty list as argument (should give empty dict):
    return_property = tango_database_test.get_property([])
    assert return_property == {}

    # With empty tuple as argument (should give empty dict):
    return_property = tango_database_test.get_property(())
    assert return_property == {}

    ## delete_property overloads:
    # With str as argument:
    property_name, _, _ = get_name_value(1)
    tango_database_test.delete_property(property_name)
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 8

    # With DbDatum as argument:
    property_name, _, _ = get_name_value(2)
    ddatum = DbDatum(property_name)
    tango_database_test.delete_property(ddatum)
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 7

    # With DbData as argument:
    property_name, _, _ = get_name_value(3)
    ddatum = DbDatum(property_name)
    ddata = DbData()
    ddata.append(ddatum)
    tango_database_test.delete_property(ddata)
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 6

    # With dict[name: value] as argument:
    property_name, _, _ = get_name_value(4)
    tango_database_test.delete_property({property_name: ""})
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 5

    # With dict[name: DaDatum] as argument:
    property_name, _, _ = get_name_value(5)
    ddatum = DbDatum(property_name)
    tango_database_test.delete_property({"some_name": ddatum})
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 4

    # With list[DaDatum] as argument:
    property_name_1, _, _ = get_name_value(6)
    ddatum_1 = DbDatum(property_name_1)
    property_name_2, _, _ = get_name_value(7)
    ddatum_2 = DbDatum(property_name_2)
    tango_database_test.delete_property([ddatum_1, ddatum_2])
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 2

    # With list[str] as argument:
    property_name_1, _, _ = get_name_value(8)
    property_name_2, _, _ = get_name_value(9)
    tango_database_test.delete_property([property_name_1, property_name_2])
    return_property_list = tango_database_test.get_property_list("*")
    assert len(return_property_list) == 0

    # With empty dict as argument (should not raise):
    tango_database_test.delete_property({})

    # With empty list as argument (should not raise):
    tango_database_test.delete_property([])

    # With empty tuple as argument (should not raise):
    tango_database_test.delete_property(())


def test_info(tango_database_test):
    info = tango_database_test.info()

    assert info.dev_class == "DataBase"
    assert info.doc_url == "Doc URL = http://www.tango-controls.org"
    assert info.server_id == "DataBaseds/2"



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

import sys
import os
import re

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.abspath("../"))
sys.path.append(os.path.abspath("./"))


# Import tango
import tango

print(f"Building documentation for PyTango {tango.Release.version_long}")
print(f"Using PyTango from: {os.path.dirname(tango.__file__)}")

needs_sphinx = "1.0"

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.imgmath",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    "sphinx_design",
    "sphinxcontrib.mermaid",
]

myst_enable_extensions = [
    "attrs_inline",
    "colon_fence",
    "deflist",
]

# when copying code blocks, exclude line numbers, prompts (like >>>), and the output
copybutton_exclude = ".linenos, .gp, .go"

autodoc_type_aliases = {"DevState": "DevState"}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# The encoding of source files.
# source_encoding = 'utf-8'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "PyTango"
# copyright in _templates/copyright.html

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ".".join(tango.Release.version.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = tango.Release.version_long

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of glob-style patterns that should be excluded when looking for source files.
# They are matched against the source file names relative to the source directory,
# using slashes as directory separators on all platforms
exclude_patterns = ["README.md"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "sphinx_book_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "logo": {
        "alt_text": "PyTango documentation - Home",
        "image_light": "_static/logo-doc.png",
        "image_dark": "_static/logo-doc.png",
    },
    "repository_url": "https://gitlab.com/tango-controls/pytango",
    "path_to_docs": "doc",
    "repository_branch": "develop",
    "use_source_button": True,
    "use_edit_page_button": True,
    "navigation_depth": 6,
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = ['_theme']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = "PyTango documentation"

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = "PyTango"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/logomark-32px.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_use_modindex = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "PyTangodoc"


# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", "PyTango.tex", "PyTango Documentation", "PyTango team", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = "_static/logo-large.png"

latex_elements = {
    "fontpkg": "\\usepackage{palatino}",
    "papersize": "a4paper",
    "pointsize": "10pt",
}
latex_show_urls = "no"

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True

# -- Options for reference to other documentation ------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "ipython": ("https://ipython.org/ipython-doc/stable", None),
    "pymongo": ("https://pymongo.readthedocs.io/en/stable", None),
    "couchdb": ("https://couchdb-python.readthedocs.io/en/latest", None),
    "pycassa": ("https://pycassa.github.io/pycassa", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/en/latest", None),
    "opentelemetry": ("https://opentelemetry-python.readthedocs.io/en/latest", None),
    "tangodoc": ("https://tango-controls.readthedocs.io/en/latest/", None),
}

todo_include_todos = True


def copy_spaces(origin):
    r = ""
    for x in range(len(origin)):
        if origin[x] in (" ", "\t"):
            r += origin[x]
        else:
            return r
    return r


def type_to_link(tipus):
    if tipus[:9] == "sequence<" and tipus[-1:] == ">":
        return "sequence<" + type_to_link(tipus[9:-1]) + ">"
    # elif tipus in dir(PyTango):
    else:
        return ":class:`" + tipus + "`"
    # else:
    #    return tipus


def type_to_pytango_link(tipus):
    if tipus[:9] == "sequence<" and tipus[-1:] == ">":
        return "sequence<" + type_to_link(tipus[9:-1]) + ">"
    elif tipus in dir(tango):
        return ":class:`" + tipus + "`"
    else:
        return tipus


def possible_type_to_link(text):
    if len(text) and text[0] == "(" and text[-1] == ")":
        return "(" + type_to_link(text[1:-1]) + ")"
    return text


def parse_typed_line(line):
    spacesSplit = line.strip().split(" ")
    first = spacesSplit[0].strip()
    return possible_type_to_link(first) + " " + " ".join(spacesSplit[1:])


def parse_parameters(line):
    spaces = copy_spaces(line)
    miniLine = line.strip()

    if miniLine[:2] != "- ":
        return line

    spl = miniLine[2:].split(":", 1)

    assert len(spl) == 2

    return spaces + ":" + spl[0].strip() + ": " + parse_typed_line(spl[1])


def parse_bullet_with_type(line):
    spaces = copy_spaces(line)
    miniLine = line.strip()

    if miniLine[:2] not in ["- ", "* "]:
        return line

    spl = miniLine.split(":", 1)

    if len(spl) != 2:
        return line

    return spaces + spl[0] + ": " + parse_typed_line(spl[1])


def parse_throws(line):
    words = re.split(r"(\W+)", line)
    assert line == "".join(words)
    return "".join(map(type_to_pytango_link, words))


def split_signature(text):
    if text is None:
        return None

    # split "fname(params)", "returntype"
    ops = text.split("->")
    if len(ops) != 2:
        return None

    # get rid of "fname"
    params = ops[0].strip()
    ret_type = ops[1].strip()
    p = params.find("(")
    if p < 0:
        return None
    params = params[p:]
    return params, ret_type


_with_only_one_signature_methods = {}


def fix_indentation(lines):
    paragraphs = []
    paragraph = []
    block_indents = []

    for line in lines:
        indent = len(line) - len(line.lstrip())
        stripped_line = line.strip()
        if stripped_line and block_indents:
            is_bullet = stripped_line.startswith("* ") or stripped_line.startswith("- ")
            is_bullet_in_same_block = is_bullet and indent == block_indents[-1]
            if indent <= block_indents[-1] and not is_bullet_in_same_block:
                block_indents.pop()
                if paragraph:
                    paragraphs.append(paragraph)
                    paragraph = []
        if stripped_line == "" and not block_indents:
            if paragraph:
                paragraphs.append(paragraph)
                paragraph = []
            paragraphs.append([""])  # preserve blank line
        else:
            paragraph.append(line)
        is_new_block = "::" in line or stripped_line.endswith(":")
        if is_new_block:
            block_indents.append(indent)

    if paragraph:
        paragraphs.append(paragraph)

    reformatted = [""]  # start with blank line
    for para in paragraphs:
        if para == [""]:
            reformatted.append("")
        else:
            min_indent = min(
                (len(line) - len(line.lstrip()) for line in para if line.strip()),
                default=0,
            )
            for line in para:
                new_trimmed_line = line[min_indent:] if line.strip() else ""
                reformatted.append(new_trimmed_line)
    lines.clear()
    lines.extend(reformatted)


def __reformat_lines(app, what, name, obj, options, lines):
    global _with_only_one_signature_methods
    if what != "method":
        for ln in range(len(lines)):
            lines[ln] = parse_bullet_with_type(lines[ln])
        return

    fix_indentation(lines)

    toinsert = []
    parsingParameters = False
    parsingThrows = False

    toinsert.append((0, ""))

    for ln in range(len(lines)):
        line = lines[ln]

        if len(line) and line[0] != " ":
            if name in _with_only_one_signature_methods:
                # This method has one and only one signature. So it will
                # be displayed by sphinx, there's no need for us to fake
                # it here...
                lines[ln] = ""
            else:
                parentesis = line.split("(", 1)
                fname = parentesis[0].strip()
                if len(parentesis) == 2 and fname == name.rsplit(".", 1)[1]:
                    sg = split_signature(line)
                    if sg is not None:
                        # Main lines are like small titles (**bold**):
                        lines[ln] = (
                            "**"
                            + fname
                            + "** *"
                            + sg[0]
                            + "* **->** "
                            + type_to_link(sg[1])
                        )
                        # Add an ENTER after the title, to make a different
                        # paragraph. So if I have 2 signatures, there's no problem
                        # with it...
                        toinsert.append((ln + 1, ""))

                    ## Main lines are like small titles (**bold**):
                    # lines[ln]='**' + line.strip() + '**'
                    ## Add an ENTER after the title, to make a different
                    ## paragraph. So if I have 2 signatures, there's no problem
                    ## with it...
                    # toinsert.append((ln+1, ""))

        # Mark the "New in this version" lines...
        if line.strip()[:14] == "New in PyTango":
            lines[ln] = copy_spaces(lines[ln]) + "*" + line.strip() + "*"
            parsingParameters = False
            parsingThrows = False

        # Look for special control_words
        # To replace the actual syntax: "Return   : something"
        # with the one understood by reStructuredText ":Return: something"
        spl = line.strip().split(":", 1)
        control_word = spl[0].strip()

        if (len(spl) != 2) or (
            control_word
            not in ["Parameters", "Return", "Throws", "Example", "See Also"]
        ):
            if parsingParameters:
                lines[ln] = parse_parameters(line)
            elif parsingThrows:
                lines[ln] = parse_throws(line)
            continue

        parsingParameters = False
        parsingThrows = False
        spaces = copy_spaces(line)

        # The Example control word is even more special. I will put
        # the contents from the following line into a code tag (::)
        if control_word == "Example":
            lines[ln] = spaces + ":" + control_word + ": " + spl[1]
            toinsert.append((ln + 1, ""))
            toinsert.append((ln + 1, spaces + " ::"))
            toinsert.append((ln + 1, ""))
        elif control_word == "Parameters":
            lines[ln] = spaces + ":Parameters:" + parse_parameters(spl[1])
            parsingParameters = True
        elif control_word == "Return":
            lines[ln] = spaces + ":Return: " + parse_typed_line(spl[1])
        elif control_word == "Throws":
            lines[ln] = spaces + ":Throws:" + parse_throws(spl[1])
            parsingThrows = True
        else:
            lines[ln] = spaces + ":" + control_word + ": " + spl[1]

    for x in range(len(toinsert) - 1, -1, -1):
        pos, txt = toinsert[x]
        lines.insert(pos, txt)


def setup(app):
    # sphinx will call this method when it finds an object to document.
    app.connect("autodoc-process-docstring", __reformat_lines)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
# Imports

import time
import sys
import gc
import inspect

from dataclasses import dataclass
from threading import Thread

import pytest
from io import StringIO

from tango import (
    EventType,
    DeviceProxy,
    DevEncoded,
    GreenMode,
    AttrQuality,
    DevFailed,
    EnsureOmniThread,
    is_omni_thread,
    EventSubMode,
    EventReason,
    EventData,
    ExtractAs,
)

from tango import Except
from tango.server import Device
from tango.server import command, attribute
from tango.test_utils import DeviceTestContext, assert_close
from tango.utils import EventCallback, PyTangoUserWarning


MAX_RETRIES = 200
DELAY_PER_RETRY = 0.05

event_results = []


async def async_callback(evt):
    event_callback(evt)


def event_callback(evt):
    if evt.err:
        event_results.append(evt.errors[0].desc)
        assert evt.attr_value is None
    elif hasattr(evt, "ctr"):
        event_results.append(evt.ctr)
    elif hasattr(evt, "attr_value"):
        event_results.append(evt.attr_value.value)
    else:
        event_results.append(evt)


# Test device
class EventDevice(Device):
    _requested_event_type = None
    _base = 0
    _slow_attr_last_read_time = 0.0

    def init_device(self):
        self.set_change_event("attr", implemented=True, detect=False)
        # even if set_alarm_event is not necessary after we did set_change_event,
        # we call it to test that function works
        self.set_alarm_event("attr", implemented=True, detect=False)
        self.set_data_ready_event("attr", implemented=True)
        self.set_archive_event("attr", implemented=True, detect=False)

        self.set_change_event("dev_encoded_attr", implemented=True, detect=False)

    @attribute(polling_period=1000)
    def attr_no_events(self):
        return 1

    @attribute()
    def attr(self) -> int:
        # to avoid sending events at subscription
        if self._requested_event_type is not None:
            self._base += 10
            self._send_events("attr", self._requested_event_type)
        return 0

    @attr.write
    def attr(self, event_type):
        self._base += 10
        self._requested_event_type = event_type
        self._send_events("attr", event_type)

    @attribute(dtype=DevEncoded)
    def dev_encoded_attr(self):
        return "a", "b"

    @attribute(change_event_implemented=True, change_event_detect=False)
    def slow_attr(self) -> int:
        time.sleep(0.1)
        self._slow_attr_last_read_time = time.time()
        time.sleep(0.1)
        return 0

    @attribute
    def slow_attr_last_read_time(self) -> float:
        return self._slow_attr_last_read_time

    @attribute(
        alarm_event_implemented=True,
        alarm_event_detect=False,
        archive_event_implemented=True,
        archive_event_detect=False,
        change_event_implemented=True,
        change_event_detect=False,
        data_ready_event_implemented=True,
    )
    def attr_decorator_with_kwords(self) -> int:
        return 0

    attr_class_with_kwords = attribute(
        alarm_event_implemented=True,
        alarm_event_detect=False,
        archive_event_implemented=True,
        archive_event_detect=False,
        change_event_implemented=True,
        change_event_detect=False,
        data_ready_event_implemented=True,
    )

    def read_attr_class_with_kwords(self) -> int:
        return 0

    @command
    def reset(self):
        self._requested_event_type = None
        self._base = 0

    def _send_events(self, attr_name: str, event_type: int):
        # to test fire_xxx_event methods we should have attr object and exception, converted to DevFailed:
        attr = self.get_device_attr().get_attr_by_name(attr_name)
        try:
            raise Exception(f"test exception {self._base + 1}")
        except Exception:
            test_exception = Except.to_dev_failed(*sys.exc_info())

        if event_type == EventType.USER_EVENT:
            self.push_event(attr_name, [], [], self._base + 1)
            self.push_event(
                attr_name, [], [], self._base + 2, 3.0, AttrQuality.ATTR_WARNING
            )
            self.push_event(
                attr_name, [], [], Exception(f"test exception {self._base}")
            )

        elif event_type == EventType.ARCHIVE_EVENT:
            self.push_archive_event(attr_name, self._base + 1)
            self.push_archive_event(
                attr_name, self._base + 2, 3.0, AttrQuality.ATTR_WARNING
            )
            self.push_archive_event(
                attr_name, Exception(f"test exception {self._base}")
            )

        elif event_type == EventType.CHANGE_EVENT:
            self.push_change_event(attr_name, self._base + 1)

            self.push_change_event(
                attr_name, self._base + 2, 3.0, AttrQuality.ATTR_WARNING
            )

            attr.set_value(self._base + 3)
            attr.fire_change_event()

            self.push_change_event(attr_name, Exception(f"test exception {self._base}"))

            attr.fire_change_event(test_exception)

        elif event_type == EventType.ALARM_EVENT:
            self.push_alarm_event(attr_name, self._base + 1)

            self.push_alarm_event(
                attr_name, self._base + 2, 3.0, AttrQuality.ATTR_WARNING
            )

            attr.set_value(self._base + 3)
            attr.fire_alarm_event()

            self.push_alarm_event(attr_name, Exception(f"test exception {self._base}"))

            attr.fire_alarm_event(test_exception)

        elif event_type == EventType.DATA_READY_EVENT:
            self.push_data_ready_event(attr_name, self._base + 1)

    @command
    def send_events(self, event_type: int):
        self._send_events("attr", event_type)

    @command
    def send_slow_event(self):
        self.push_change_event("slow_attr", 42)

    @command
    def send_event_no_data(self, attr_name: str):
        self.push_event(attr_name, [], [])

    @command
    def send_archive_event_no_data(self, attr_name: str):
        self.push_archive_event(attr_name)

    @command
    def send_change_event_no_data(self, attr_name: str):
        self.push_change_event(attr_name)

    @command
    def send_alarm_event_no_data(self, attr_name: str):
        self.push_alarm_event(attr_name)

    @command
    def send_change_event_with_dim_argument(self):
        dim_x_dim_y_unsupported = "dim_x and dim_y arguments are no longer supported"
        now = time.time()

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("attr", 42, 1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("attr", 42, 1, 1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("attr", 42, dim_x=1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("attr", 42, dim_x=1, dim_y=1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("attr", 42, now, AttrQuality.ATTR_VALID, 1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("attr", 42, now, AttrQuality.ATTR_VALID, 1, 1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("dev_encoded_attr", "a", "b", 1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("dev_encoded_attr", "a", "b", 1, 1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("dev_encoded_attr", "a", "b", dim_x=1)

        with pytest.raises(TypeError, match=dim_x_dim_y_unsupported):
            self.push_change_event("dev_encoded_attr", "a", "b", dim_x=1, dim_y=1)

        # check some valid dev encoded events work
        self.push_change_event("dev_encoded_attr", "a", "b")
        self.push_change_event(
            "dev_encoded_attr", "a", "b", now, AttrQuality.ATTR_VALID
        )

    @command(dtype_in=str)
    def add_dyn_attr(self, name):
        attr = attribute(name=name, dtype="float", fget=self.read)
        self.add_attribute(attr)

    @command(dtype_in=str)
    def delete_dyn_attr(self, name):
        self.remove_attribute(name)

    def read(self, attr):
        attr.set_value(1.23)


cmd_list = {
    "Init",
    "State",
    "Status",
    "reset",
    "add_dyn_attr",
    "delete_dyn_attr",
    "send_events",
    "send_event_no_data",
    "send_slow_event",
    "send_archive_event_no_data",
    "send_change_event_no_data",
    "send_alarm_event_no_data",
    "send_change_event_with_dim_argument",
}

attr_list = {
    "State",
    "Status",
    "attr",
    "attr_class_with_kwords",
    "attr_decorator_with_kwords",
    "attr_no_events",
    "dev_encoded_attr",
    "slow_attr",
    "slow_attr_last_read_time",
}

DEVICE_NAME = "test/nodb/eventdevice"


# Device fixture
@pytest.fixture(scope="module")
def event_device_context():
    context = DeviceTestContext(
        EventDevice, device_name=DEVICE_NAME, host="127.0.0.1", process=True
    )
    with context:
        yield context


@pytest.fixture(scope="module")
def event_device_with_green_modes(event_device_context, green_mode_device_proxy):
    return green_mode_device_proxy(event_device_context.get_device_access())


@pytest.fixture(scope="module")
def event_device(event_device_context):
    return DeviceProxy(event_device_context.get_device_access())


# Tests
def assert_events_received(proxy, expected_res):
    for retry_count in range(MAX_RETRIES):
        proxy.read_attribute("state", wait=True)
        if len(event_results) >= len(expected_res):
            assert_close(event_results, expected_res)
            return
        time.sleep(DELAY_PER_RETRY)
    timeout_seconds = MAX_RETRIES * DELAY_PER_RETRY
    pytest.fail(
        f"Timeout, waiting for event, after {timeout_seconds} sec over {MAX_RETRIES} retries. "
        f"Occasionally happens, probably due to CI test runtime environment"
    )


def run_event_test(proxy, event_type, cb, expected_res):
    event_results.clear()
    proxy.command_inout("reset", wait=True)

    eid_change = proxy.subscribe_event("attr", event_type, cb, wait=True)

    # Trigger events from command
    proxy.command_inout("send_events", event_type, wait=True)

    # Trigger events from attribute write method
    proxy.write_attribute("attr", event_type, wait=True)

    # Trigger events from attribute read method
    proxy.read_attribute("attr", wait=True)

    # Test the event values
    assert_events_received(proxy, expected_res)

    # Unsubscribe
    proxy.unsubscribe_event(eid_change)


def test_attribute_with_no_events_cannot_be_subscribed(event_device):
    with pytest.raises(
        DevFailed,
        match=r"Event properties \(abs_change or rel_change\) "
        r"for attribute attr_no_events are not set",
    ):
        _ = event_device.subscribe_event(
            "attr_no_events", EventType.CHANGE_EVENT, EventCallback()
        )

    with pytest.raises(
        DevFailed,
        match=r"Archive event properties \(archive_abs_change or archive_rel_change or archive_period\) "
        r"for attribute attr_no_events are not set",
    ):
        _ = event_device.subscribe_event(
            "attr_no_events", EventType.ARCHIVE_EVENT, EventCallback()
        )


def test_old_push_event_signatures_no_longer_supported(event_device):
    event_device.send_change_event_with_dim_argument()


def test_change_event(event_device_with_green_modes):
    expected_res = [
        0,
        1,
        2,
        3,
        "Exception: test exception 0\n",
        "Exception: test exception 1\n",
        11,
        12,
        13,
        "Exception: test exception 10\n",
        "Exception: test exception 11\n",
        21,
        22,
        23,
        "Exception: test exception 20\n",
        "Exception: test exception 21\n",
    ]

    if event_device_with_green_modes.get_green_mode() == GreenMode.Asyncio:
        run_event_test(
            event_device_with_green_modes,
            EventType.CHANGE_EVENT,
            async_callback,
            expected_res,
        )
        with pytest.warns(DeprecationWarning):
            run_event_test(
                event_device_with_green_modes,
                EventType.CHANGE_EVENT,
                event_callback,
                expected_res,
            )
    else:
        run_event_test(
            event_device_with_green_modes,
            EventType.CHANGE_EVENT,
            event_callback,
            expected_res,
        )


def test_alarm_event(event_device):
    expected_res = [
        0,
        1,
        2,
        3,
        "Exception: test exception 0\n",
        "Exception: test exception 1\n",
        11,
        12,
        13,
        "Exception: test exception 10\n",
        "Exception: test exception 11\n",
        21,
        22,
        23,
        "Exception: test exception 20\n",
        "Exception: test exception 21\n",
    ]

    run_event_test(
        event_device,
        EventType.ALARM_EVENT,
        event_callback,
        expected_res,
    )


def test_user_event(event_device):
    expected_res = [
        0,
        1,
        2,
        "Exception: test exception 0\n",
        11,
        12,
        "Exception: test exception 10\n",
        21,
        22,
        "Exception: test exception 20\n",
    ]

    run_event_test(
        event_device,
        EventType.USER_EVENT,
        event_callback,
        expected_res,
    )


def test_archive_event(event_device):
    expected_res = [
        0,
        1,
        2,
        "Exception: test exception 0\n",
        11,
        12,
        "Exception: test exception 10\n",
        21,
        22,
        "Exception: test exception 20\n",
    ]

    run_event_test(
        event_device,
        EventType.ARCHIVE_EVENT,
        event_callback,
        expected_res,
    )


def test_subscribe_data_ready_event(event_device):
    expected_res = [1, 11, 21]
    run_event_test(
        event_device,
        EventType.DATA_READY_EVENT,
        event_callback,
        expected_res,
    )


def test_subscribe_interface_event(event_device):

    def cb(event):
        if len(event_results) == 0:
            assert {cmd.cmd_name for cmd in event.cmd_list} == cmd_list
            assert {att.name for att in event.att_list} == attr_list
        elif len(event_results) == 1:
            assert {cmd.cmd_name for cmd in event.cmd_list} == cmd_list
            assert {att.name for att in event.att_list} == attr_list | {"bla"}
        else:
            assert {cmd.cmd_name for cmd in event.cmd_list} == cmd_list
            assert {att.name for att in event.att_list} == attr_list

        event_results.append(True)

    event_results.clear()

    # Subscribe
    eid = event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, cb, wait=True)
    # Trigger an event
    event_device.command_inout("add_dyn_attr", "bla", wait=True)
    event_device.read_attribute("bla", wait=True)
    # Wait for tango event
    assert_events_received(event_device, [True, True])

    event_device.command_inout("delete_dyn_attr", "bla", wait=True)
    # Wait for tango event
    assert_events_received(event_device, [True, True, True])
    # Unsubscribe
    event_device.unsubscribe_event(eid)


def test_push_event_with_event_callback(event_device):
    string = StringIO()
    cb = EventCallback(fd=string)

    # to reduce tests amount here we test only change event
    eid = event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, wait=True)
    # trigger an event
    event_device.command_inout("send_events", EventType.CHANGE_EVENT, wait=True)
    # wait for tango event
    for retry_count in range(MAX_RETRIES):
        event_device.read_attribute("state", wait=True)
        if len(cb.get_events()) > 5:
            break
        time.sleep(DELAY_PER_RETRY)
    if retry_count + 1 >= MAX_RETRIES:
        timeout_seconds = retry_count * DELAY_PER_RETRY
        pytest.fail(
            f"Timeout, waiting for event, after {timeout_seconds}sec over {MAX_RETRIES} retries. "
            f"Occasionally happens, probably due to CI test runtime environment"
        )
    # Test the event values and timestamp
    events = cb.get_events()
    results = [evt.attr_value.value for evt in events[:4]]
    assert results == [0, 1, 2, 3]
    assert events[2].attr_value.time.totime() == 3.0
    assert events[4].errors[0].desc == "Exception: test exception 0\n"
    assert events[5].errors[0].desc == "Exception: test exception 1\n"
    for evt in events:
        assert evt.device is event_device

    # Check string
    for line in [
        "TEST/NODB/EVENTDEVICE ATTR CHANGE SUBSUCCESS [ATTR_VALID] 0",
        "TEST/NODB/EVENTDEVICE ATTR CHANGE UPDATE [ATTR_VALID] 1",
        "TEST/NODB/EVENTDEVICE ATTR CHANGE UPDATE [ATTR_WARNING] 2",
        "TEST/NODB/EVENTDEVICE ATTR CHANGE UPDATE [ATTR_VALID] 3",
        "TEST/NODB/EVENTDEVICE ATTR CHANGE UPDATE [PyDs_PythonError] Exception: test exception 0",
        "TEST/NODB/EVENTDEVICE ATTR CHANGE UPDATE [PyDs_PythonError] Exception: test exception 1",
    ]:
        assert line in string.getvalue()
    # Unsubscribe
    event_device.unsubscribe_event(eid)


def test_send_events_no_data(event_device):
    event_device.command_inout("send_event_no_data", "state", wait=True)
    event_device.command_inout("send_archive_event_no_data", "State", wait=True)
    event_device.command_inout("send_change_event_no_data", "STaTe", wait=True)
    event_device.command_inout("send_alarm_event_no_data", "state", wait=True)

    event_device.command_inout("send_event_no_data", "status", wait=True)
    event_device.command_inout("send_archive_event_no_data", "Status", wait=True)
    event_device.command_inout("send_change_event_no_data", "STaTus", wait=True)
    event_device.command_inout("send_alarm_event_no_data", "status", wait=True)

    expected_err_msg = (
        "Cannot push event for attribute attr without data. "
        "Pushing event without data parameter is only allowed for State and Status attributes."
    )

    with pytest.raises(DevFailed, match=expected_err_msg):
        event_device.command_inout("send_event_no_data", "attr", wait=True)

    with pytest.raises(DevFailed, match=expected_err_msg):
        event_device.command_inout("send_archive_event_no_data", "attr", wait=True)

    with pytest.raises(DevFailed, match=expected_err_msg):
        event_device.command_inout("send_change_event_no_data", "attr", wait=True)

    with pytest.raises(DevFailed, match=expected_err_msg):
        event_device.command_inout("send_alarm_event_no_data", "attr", wait=True)


def test_main_thread_is_omni_thread():
    assert is_omni_thread()


def test_ensure_omni_thread_main_thread_is_omni_thread():
    with EnsureOmniThread():
        assert is_omni_thread()


def test_user_thread_is_not_omni_thread():
    thread_is_omni = dict(result=None)  # use a dict so thread can modify it

    def thread_func():
        thread_is_omni["result"] = is_omni_thread()

    thread = Thread(target=thread_func)
    thread.start()
    thread.join()
    assert thread_is_omni["result"] is False


def test_ensure_omni_thread_user_thread_is_omni_thread():
    thread_is_omni = dict(result=None)  # use a dict so thread can modify it

    def thread_func():
        with EnsureOmniThread():
            thread_is_omni["result"] = is_omni_thread()

    thread = Thread(target=thread_func)
    thread.start()
    thread.join()
    assert thread_is_omni["result"] is True


def test_subscribe_change_event_from_user_thread(event_device):
    event_results.clear()

    def thread_func():
        with EnsureOmniThread():
            eid = event_device.subscribe_event(
                "attr", EventType.CHANGE_EVENT, event_callback, wait=True
            )
            while running:
                time.sleep(DELAY_PER_RETRY)
            event_device.unsubscribe_event(eid)

    # Start the thread
    thread = Thread(target=thread_func)
    running = True
    thread.start()
    # Wait for tango events
    for retry_count in range(MAX_RETRIES):
        event_device.read_attribute("state", wait=True)
        if len(event_results) == 1:
            # Trigger an event (1 result means thread has completed subscription,
            # as that results in an initial callback)
            event_device.command_inout("send_events", EventType.CHANGE_EVENT, wait=True)
        elif len(event_results) > 2:
            # At least 2 events means an event was received after subscription
            break
        time.sleep(DELAY_PER_RETRY)
    # Stop the thread
    running = False

    thread.join()
    if retry_count + 1 >= MAX_RETRIES:
        timeout_seconds = retry_count * DELAY_PER_RETRY
        pytest.fail(
            f"Timeout, waiting for event, after {timeout_seconds}sec over {MAX_RETRIES} retries. "
            f"Occasionally happens, probably due to CI test runtime environment"
        )
    # Test the event values
    assert event_results == [
        0,
        1,
        2,
        3,
        "Exception: test exception 0\n",
        "Exception: test exception 1\n",
    ]


def test_get_events(event_device):
    for receiver in ["cb", "no"]:
        for event_type in EventType.values.values():
            event_results.clear()

            if event_type == EventType.PERIODIC_EVENT:  # needs polling, so skip
                continue
            eid = event_device.subscribe_event("attr", event_type, 1, wait=True)

            # DATA_READY_EVENT does not send automatically at the subscription
            if event_type == EventType.DATA_READY_EVENT:
                event_device.send_events(event_type, wait=True)

            for retry_count in range(MAX_RETRIES):
                if receiver == "no":
                    if len(event_device.get_events(eid)):
                        break
                else:
                    event_device.get_events(eid, event_callback)
                    if len(event_results):
                        break
                time.sleep(DELAY_PER_RETRY)

            if retry_count + 1 >= MAX_RETRIES:
                timeout_seconds = retry_count * DELAY_PER_RETRY
                pytest.fail(
                    f"Timeout, waiting for event, after {timeout_seconds}sec over {MAX_RETRIES} retries. "
                    f"Occasionally happens, probably due to CI test runtime environment"
                )

            event_device.unsubscribe_event(eid)


def test_callback_was_dereferenced(event_device):

    def fcb(_event):
        pass

    initial_refs = len(gc.get_referrers(fcb))
    ids = [
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, fcb, wait=True)
        for _ in range(10)
    ]
    assert initial_refs < len(gc.get_referrers(fcb))

    for id in ids:
        event_device.unsubscribe_event(id, wait=True)

    assert initial_refs == len(gc.get_referrers(fcb))


def test_event_after_unsubscription_crash(event_device):

    class DummyCallback:

        def push_event(self, event):
            pass

    class MainCallback:

        def __init__(self):
            self.device = DeviceProxy(DEVICE_NAME)
            self.eid = self.device.subscribe_event(
                "attr", EventType.CHANGE_EVENT, DummyCallback()
            )
            self.is_first_time = True

        def push_event(self, _event):
            if self.is_first_time:
                # callback from initial subscription
                self.is_first_time = False
            elif self.device:
                # callback from a change event, and haven't unsubscribed yet
                self.device.unsubscribe_event(self.eid)
                self.device = None  # remove reference, so it can be deleted
            return

    for _ in range(50):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, MainCallback())
        event_device.send_events(EventType.CHANGE_EVENT)


def test_events_enabled_by_kwords(event_device):
    """
    In this test we only check, that we can subscribe, i.e., that out kwords were passed to Tango::Attr
    """
    events_to_check = [
        EventType.ALARM_EVENT,
        EventType.ARCHIVE_EVENT,
        EventType.CHANGE_EVENT,
        EventType.DATA_READY_EVENT,
    ]

    for attr_name in ["attr_decorator_with_kwords", "attr_class_with_kwords"]:
        eids = [
            event_device.subscribe_event(
                attr_name, event_type, event_callback, wait=True
            )
            for event_type in events_to_check
        ]
        for eid in eids:
            event_device.unsubscribe_event(eid)


@dataclass
class SubscriptionModeExpectation:
    sub_mode: EventSubMode
    read_during_subscription: bool
    read_after_subscription: bool
    initial_event_after_subscription: bool
    initial_event_has_data: bool
    exception_on_subscription_failure: bool

    def __str__(self):
        return str(self.sub_mode)


subscription_mode_tests = (
    SubscriptionModeExpectation(
        sub_mode=EventSubMode.Sync,
        read_during_subscription=False,
        read_after_subscription=False,
        initial_event_after_subscription=False,
        initial_event_has_data=False,
        exception_on_subscription_failure=True,
    ),
    SubscriptionModeExpectation(
        sub_mode=EventSubMode.SyncRead,
        read_during_subscription=True,
        read_after_subscription=False,
        initial_event_after_subscription=True,
        initial_event_has_data=True,
        exception_on_subscription_failure=True,
    ),
    SubscriptionModeExpectation(
        sub_mode=EventSubMode.Async,
        read_during_subscription=False,
        read_after_subscription=False,
        initial_event_after_subscription=True,
        initial_event_has_data=False,
        exception_on_subscription_failure=False,
    ),
    SubscriptionModeExpectation(
        sub_mode=EventSubMode.AsyncRead,
        read_during_subscription=False,
        read_after_subscription=True,
        initial_event_after_subscription=True,
        initial_event_has_data=True,
        exception_on_subscription_failure=False,
    ),
    SubscriptionModeExpectation(
        sub_mode=EventSubMode.Stateless,
        read_during_subscription=True,
        read_after_subscription=False,
        initial_event_after_subscription=True,
        initial_event_has_data=True,
        exception_on_subscription_failure=False,
    ),
)


def wait_for_single_event(proxy, events, timeout) -> EventData:
    t0 = time.time()
    while time.time() - t0 < timeout:
        proxy.read_attribute("state", wait=True)
        if len(events) == 1:
            break
        time.sleep(DELAY_PER_RETRY)
    assert len(events) == 1
    return events[0]


@pytest.mark.parametrize("subscription_mode_tests", subscription_mode_tests, ids=str)
def test_event_subscription_sub_modes(subscription_mode_tests):
    params = subscription_mode_tests

    events = []

    with DeviceTestContext(EventDevice, process=True) as event_device:
        start_time = time.time()
        eid = event_device.subscribe_event(
            "slow_attr",
            EventType.CHANGE_EVENT,
            cb=events.append,
            sub_mode=params.sub_mode,
        )
        subscription_completed_time = time.time()

        if params.initial_event_after_subscription:
            event = wait_for_single_event(event_device, events, timeout=1.0)
            assert event.event_reason == EventReason.SubSuccess
            if params.initial_event_has_data:
                assert event.attr_value is not None
            else:
                assert event.attr_value is None

        last_read_time = event_device.slow_attr_last_read_time
        if params.read_during_subscription:
            assert start_time < last_read_time < subscription_completed_time
        elif params.read_after_subscription:
            assert start_time < subscription_completed_time < last_read_time
        else:
            assert last_read_time == 0.0

        events.clear()
        event_device.send_slow_event()
        event = wait_for_single_event(event_device, events, timeout=0.5)
        assert event.event_reason == EventReason.Update
        assert event.attr_value is not None

        event_device.unsubscribe_event(eid)


@pytest.mark.parametrize("subscription_mode_tests", subscription_mode_tests, ids=str)
def test_event_subscription_sub_modes_invalid_attribute_fails(subscription_mode_tests):
    params = subscription_mode_tests

    events = []

    with DeviceTestContext(Device, process=True) as event_device:
        if params.exception_on_subscription_failure:
            with pytest.raises(DevFailed, match="API_AttrNotFound"):
                _ = event_device.subscribe_event(
                    "non_existent_attr",
                    EventType.CHANGE_EVENT,
                    cb=events.append,
                    sub_mode=params.sub_mode,
                )
        else:
            eid = event_device.subscribe_event(
                "non_existent_attr",
                EventType.CHANGE_EVENT,
                cb=events.append,
                sub_mode=params.sub_mode,
            )

            event = wait_for_single_event(event_device, events, 0.5)
            assert event.err
            assert len(event.errors) > 0
            assert event.event_reason == EventReason.SubFail
            assert event.attr_value is None

            event_device.unsubscribe_event(eid)


def test_subscribe_event_signatures():
    """
    Signatures to be supported:

        subscribe_event(self,            event_type, cb,        sub_mode=EventSubMode.SyncRead, *, __GREEN_KWARGS__) -> int
        subscribe_event(self,            event_type, cb,        stateless=False,                *, __GREEN_KWARGS__) -> int

        subscribe_event(self, attr_name, event_type, cb,        sub_mode=EventSubMode.SyncRead, extract_as=ExtractAs.Numpy, *, __GREEN_KWARGS__) -> int
        subscribe_event(self, attr_name, event_type, queuesize, sub_mode=EventSubMode.SyncRead, extract_as=ExtractAs.Numpy, *, __GREEN_KWARGS__) -> int
        subscribe_event(self, attr_name, event_type, cb,        filters=[], stateless=False,    extract_as=ExtractAs.Numpy, *, __GREEN_KWARGS__) -> int
        subscribe_event(self, attr_name, event_type, queuesize, filters=[], stateless=False,    extract_as=ExtractAs.Numpy, *, __GREEN_KWARGS__) -> int
    """

    last_called_method_and_args = []

    def cb(evnt):
        pass

    cb_with_push_event = EventCallback()

    def subscription_call_mock(method, *args):
        last_called_method_and_args.append((method, *args))
        return len(last_called_method_and_args)

    def check_last_called_method(*method_and_args):
        for called, expected in zip(last_called_method_and_args[-1], method_and_args):
            if callable(expected):
                callback = inspect.unwrap(called.push_event)
                assert callback == expected
            else:
                assert called == expected

    # functionality is mocked, so we don't need a real device
    event_device = DeviceProxy("tango://127.0.0.1:0/dummy/test/device#dbase=no")

    with pytest.warns(PyTangoUserWarning):
        event_device.unfreeze_dynamic_interface()

    event_device.__subscribe_event_global_with_stateless_flag = (
        lambda *args: subscription_call_mock(
            "__subscribe_event_global_with_stateless_flag", *args
        )
    )
    event_device.__subscribe_event_global_with_sub_mode = (
        lambda *args: subscription_call_mock(
            "__subscribe_event_global_with_sub_mode", *args
        )
    )
    event_device.__subscribe_event_attrib_with_stateless_flag = (
        lambda *args: subscription_call_mock(
            "__subscribe_event_attrib_with_stateless_flag", *args
        )
    )
    event_device.__subscribe_event_attrib_with_sub_mode = (
        lambda *args: subscription_call_mock(
            "__subscribe_event_attrib_with_sub_mode", *args
        )
    )

    with pytest.raises(
        TypeError, match="This method is only for Interface Change Events"
    ):
        event_device.subscribe_event(EventType.CHANGE_EVENT)

    with pytest.raises(IndexError, match="Expected parameter 'cb'"):
        event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT)

    event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, cb)
    check_last_called_method(
        "__subscribe_event_global_with_sub_mode",
        EventType.INTERFACE_CHANGE_EVENT,
        cb,
        EventSubMode.SyncRead,
    )
    event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, cb=cb)
    check_last_called_method(
        "__subscribe_event_global_with_sub_mode",
        EventType.INTERFACE_CHANGE_EVENT,
        cb,
        EventSubMode.SyncRead,
    )
    event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, cb_with_push_event)
    check_last_called_method(
        "__subscribe_event_global_with_sub_mode",
        EventType.INTERFACE_CHANGE_EVENT,
        cb_with_push_event.push_event,
        EventSubMode.SyncRead,
    )
    event_device.subscribe_event(
        EventType.INTERFACE_CHANGE_EVENT, cb=cb_with_push_event
    )
    check_last_called_method(
        "__subscribe_event_global_with_sub_mode",
        EventType.INTERFACE_CHANGE_EVENT,
        cb_with_push_event.push_event,
        EventSubMode.SyncRead,
    )

    with pytest.raises(TypeError, match="Parameter 'cb' must be callable object"):
        event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, cb=None)

    with pytest.raises(TypeError, match="Parameter 'cb' must be callable object"):
        event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, None)

    with pytest.raises(
        IndexError,
        match="Expected parameter 'cb' as either positional arg 2, ",
    ):
        event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT)

    event_device.subscribe_event(
        EventType.INTERFACE_CHANGE_EVENT, cb, EventSubMode.AsyncRead
    )
    check_last_called_method(
        "__subscribe_event_global_with_sub_mode",
        EventType.INTERFACE_CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
    )
    event_device.subscribe_event(
        EventType.INTERFACE_CHANGE_EVENT, cb, sub_mode=EventSubMode.AsyncRead
    )
    check_last_called_method(
        "__subscribe_event_global_with_sub_mode",
        EventType.INTERFACE_CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
    )

    with pytest.warns(
        DeprecationWarning, match="The 'stateless' parameter is deprecated"
    ):
        event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, cb, False)

    check_last_called_method(
        "__subscribe_event_global_with_stateless_flag",
        EventType.INTERFACE_CHANGE_EVENT,
        cb,
        False,
    )
    with pytest.warns(
        DeprecationWarning, match="The 'stateless' parameter is deprecated"
    ):
        event_device.subscribe_event(
            EventType.INTERFACE_CHANGE_EVENT, cb, stateless=False
        )
    check_last_called_method(
        "__subscribe_event_global_with_stateless_flag",
        EventType.INTERFACE_CHANGE_EVENT,
        cb,
        False,
    )

    with pytest.raises(TypeError, match="Invalid type for parameter"):
        event_device.subscribe_event(EventType.INTERFACE_CHANGE_EVENT, cb, None)

    with pytest.raises(TypeError, match="Invalid type for parameter"):
        event_device.subscribe_event(
            EventType.INTERFACE_CHANGE_EVENT, cb, sub_mode=None
        )

    with pytest.raises(TypeError, match="Invalid type for parameter"):
        event_device.subscribe_event(
            EventType.INTERFACE_CHANGE_EVENT, cb, stateless=None
        )

    with pytest.raises(IndexError, match="Expected parameter 'event_type'"):
        event_device.subscribe_event("attr")

    with pytest.raises(TypeError, match="Invalid type for parameter"):
        event_device.subscribe_event("attr", None)

    with pytest.raises(TypeError, match="Invalid type for parameter"):
        event_device.subscribe_event("attr", event_type=None)

    with pytest.raises(
        TypeError, match="Either parameter 'queuesize' .* or parameter 'cb'"
    ):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT)

    event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb)
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.SyncRead,
        ExtractAs.Numpy,
    )
    event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb=cb)
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.SyncRead,
        ExtractAs.Numpy,
    )
    event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb_with_push_event)
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb_with_push_event.push_event,
        EventSubMode.SyncRead,
        ExtractAs.Numpy,
    )
    event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb=cb_with_push_event)
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb_with_push_event.push_event,
        EventSubMode.SyncRead,
        ExtractAs.Numpy,
    )

    event_device.subscribe_event("attr", EventType.CHANGE_EVENT, 1)
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        1,
        EventSubMode.SyncRead,
        ExtractAs.Numpy,
    )
    event_device.subscribe_event("attr", EventType.CHANGE_EVENT, queuesize=1)
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        1,
        EventSubMode.SyncRead,
        ExtractAs.Numpy,
    )

    with pytest.raises(
        TypeError, match="Either parameter 'queuesize' .* or parameter 'cb'"
    ):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, None)

    with pytest.raises(TypeError, match="'cb' and 'queuesize' cannot be used together"):
        event_device.subscribe_event(
            "attr",
            EventType.CHANGE_EVENT,
            cb=cb,
            queuesize=1,
        )

    with pytest.raises(TypeError, match="Invalid type for parameter 'cb=None'"):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb=None)

    with pytest.raises(TypeError, match="Invalid type for parameter 'queuesize="):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, queuesize=None)

    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, [])
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        False,
        ExtractAs.Numpy,
        [],
    )
    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, filters=[])
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        False,
        ExtractAs.Numpy,
        [],
    )

    with pytest.raises(TypeError, match="Invalid type for parameter 'filters="):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, filters=None)

    event_device.subscribe_event(
        "attr", EventType.CHANGE_EVENT, cb, EventSubMode.AsyncRead
    )
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
        ExtractAs.Numpy,
    )
    event_device.subscribe_event(
        "attr", EventType.CHANGE_EVENT, cb, sub_mode=EventSubMode.AsyncRead
    )
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
        ExtractAs.Numpy,
    )
    with pytest.raises(TypeError, match="Invalid type for parameter 'sub_mode"):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, sub_mode=None)
    with pytest.raises(TypeError, match="Invalid type for parameter 'stateless"):
        event_device.subscribe_event(
            "attr", EventType.CHANGE_EVENT, cb, [], EventSubMode.SyncRead
        )
    with pytest.raises(
        TypeError, match="'sub_mode' and 'filters' cannot be used together"
    ):
        event_device.subscribe_event(
            "attr",
            EventType.CHANGE_EVENT,
            cb,
            filters=[],
            sub_mode=EventSubMode.SyncRead,
        )
    with pytest.raises(
        TypeError, match="'sub_mode' and 'stateless' cannot be used together"
    ):
        event_device.subscribe_event(
            "attr",
            EventType.CHANGE_EVENT,
            cb,
            stateless=False,
            sub_mode=EventSubMode.SyncRead,
        )

    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, [], True)
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        True,
        ExtractAs.Numpy,
        [],
    )
    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event(
            "attr", EventType.CHANGE_EVENT, cb, [], stateless=True
        )
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        True,
        ExtractAs.Numpy,
        [],
    )
    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event(
            "attr", EventType.CHANGE_EVENT, cb, filters=["a", "b"], stateless=True
        )
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        True,
        ExtractAs.Numpy,
        ["a", "b"],
    )
    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, stateless=True)
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        True,
        ExtractAs.Numpy,
        [],
    )
    with pytest.raises(TypeError, match="Invalid type for parameter 'sub_mode"):
        event_device.subscribe_event("attr", EventType.CHANGE_EVENT, cb, False)
    with pytest.raises(TypeError, match="Invalid type for parameter 'stateless"):
        event_device.subscribe_event(
            "attr", EventType.CHANGE_EVENT, cb, [], stateless=None
        )

    event_device.subscribe_event(
        "attr", EventType.CHANGE_EVENT, cb, EventSubMode.AsyncRead, ExtractAs.Tuple
    )
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
        ExtractAs.Tuple,
    )
    event_device.subscribe_event(
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
        extract_as=ExtractAs.Tuple,
    )
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
        ExtractAs.Tuple,
    )
    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event(
            "attr",
            EventType.CHANGE_EVENT,
            cb,
            [],
            True,
            ExtractAs.Tuple,
        )
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        True,
        ExtractAs.Tuple,
        [],
    )
    with pytest.warns(
        DeprecationWarning,
        match="The 'stateless' and 'filters' parameters are deprecated",
    ):
        event_device.subscribe_event(
            "attr",
            EventType.CHANGE_EVENT,
            cb,
            [],
            True,
            extract_as=ExtractAs.Tuple,
        )
    check_last_called_method(
        "__subscribe_event_attrib_with_stateless_flag",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        True,
        ExtractAs.Tuple,
        [],
    )

    with pytest.raises(
        TypeError, match="Invalid type for parameter 'extract_as' at position 4"
    ):
        event_device.subscribe_event(
            "attr", EventType.CHANGE_EVENT, cb, EventSubMode.SyncRead, None
        )

    with pytest.raises(TypeError, match="Invalid type for parameter 'extract_as"):
        event_device.subscribe_event(
            "attr",
            EventType.CHANGE_EVENT,
            cb,
            EventSubMode.SyncRead,
            extract_as=None,
        )

    event_device.subscribe_event(
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
        ExtractAs.Tuple,
        green_mode=GreenMode.Synchronous,
    )
    check_last_called_method(
        "__subscribe_event_attrib_with_sub_mode",
        "attr",
        EventType.CHANGE_EVENT,
        cb,
        EventSubMode.AsyncRead,
        ExtractAs.Tuple,
    )

    with pytest.raises(
        TypeError,
        match="Got unexpected keyword argument.*submode.*\nAllowed.*sub_mode.*",
    ):
        event_device.subscribe_event(
            "attr",
            EventType.CHANGE_EVENT,
            cb,
            submode=EventSubMode.SyncRead,
            extract_as=None,
        )



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""This module exposes a gevent version of :class:`tango.DeviceProxy` and
:class:`tango.AttributeProxy"""

from functools import partial
from packaging.version import Version

from tango import GreenMode
from tango.device_proxy import get_device_proxy
from tango.attribute_proxy import get_attribute_proxy

__all__ = ("DeviceProxy", "AttributeProxy")


def check_requirements():
    try:
        import gevent
    except ImportError:
        raise ImportError(
            "No module named gevent. You need to install "
            "gevent module to have access to PyTango gevent "
            "green mode. Consider using the futures green mode "
            "instead"
        )

    if Version(gevent.__version__) < Version("1.0"):
        raise ImportError(
            f"You need gevent >= 1.0. You are using {gevent.__version__}. "
            f"Consider using the futures green mode instead"
        )


check_requirements()

DeviceProxy = partial(get_device_proxy, green_mode=GreenMode.Gevent)
DeviceProxy.__doc__ = """
    DeviceProxy(self, dev_name, wait=True, timeout=True) -> DeviceProxy
    DeviceProxy(self, dev_name, need_check_acc, wait=True, timeout=True) -> DeviceProxy

    Creates a *gevent* enabled :class:`~tango.DeviceProxy`.

    The DeviceProxy constructor internally makes some network calls which makes
    it *slow*. By using the gevent *green mode* you are allowing other python
    code to be executed in a cooperative way.

    .. note::
        The timeout parameter has no relation with the tango device client side
        timeout (gettable by :meth:`~tango.DeviceProxy.get_timeout_millis` and
        settable through :meth:`~tango.DeviceProxy.set_timeout_millis`)

    :param dev_name: the device name or alias
    :type dev_name: str
    :param need_check_acc: in first version of the function it defaults to True.
                           Determines if at creation time of DeviceProxy it should check
                           for channel access (rarely used)
    :type need_check_acc: bool
    :param wait: whether or not to wait for result of creating a DeviceProxy.
    :type wait: bool
    :param timeout: The number of seconds to wait for the result.
                    If None, then there is no limit on the wait time.
                    Ignored when wait is False.
    :type timeout: float
    :returns:
        if wait is True:
            :class:`~tango.DeviceProxy`
        else:
            :class:`gevent.event.AsynchResult`
    :throws:
        * a *DevFailed* if wait is True and there is an error creating
          the device.
        * a *gevent.timeout.Timeout* if wait is False, timeout is not None and
          the time to create the device has expired.

    New in PyTango 8.1.0
"""

AttributeProxy = partial(get_attribute_proxy, green_mode=GreenMode.Gevent)
AttributeProxy.__doc__ = """
    AttributeProxy(self, full_attr_name, wait=True, timeout=True) -> AttributeProxy
    AttributeProxy(self, device_proxy, attr_name, wait=True, timeout=True) -> AttributeProxy

    Creates a *gevent* enabled :class:`~tango.AttributeProxy`.

    The AttributeProxy constructor internally makes some network calls which
    makes it *slow*. By using the *gevent mode* you are allowing other python
    code to be executed in a cooperative way.

    :param full_attr_name: the full name of the attribute
    :type full_attr_name: str
    :param device_proxy: the :class:`~tango.DeviceProxy`
    :type device_proxy: DeviceProxy
    :param attr_name: attribute name for the given device proxy
    :type attr_name: str
    :param wait: whether or not to wait for result of creating an
                 AttributeProxy.
    :type wait: bool
    :param timeout: The number of seconds to wait for the result.
                    If None, then there is no limit on the wait time.
                    Ignored when wait is False.
    :type timeout: float
    :returns:
        if wait is True:
            :class:`~tango.AttributeProxy`
        else:
            :class:`gevent.event.AsynchResult`
    :throws:
        * a *DevFailed* if wait is True  and there is an error creating the
          attribute.
        * a *gevent.timeout.Timeout* if wait is False, timeout is not None
          and the time to create the attribute has expired.

    New in PyTango 8.1.0
"""

Device = DeviceProxy
Attribute = AttributeProxy

del GreenMode
del get_device_proxy
del get_attribute_proxy



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later


# Imports
import os
import inspect
import textwrap
import re

from functools import wraps

from threading import get_ident

# Tango imports
from tango._tango import GreenMode
from tango.utils import _forcefully_traced_method

__all__ = (
    "get_green_mode",
    "set_green_mode",
    "green",
    "green_callback",
    "get_executor",
    "get_object_executor",
    "switch_existing_global_executors_to_thread",
)

try:
    import gevent

    del gevent
    _gevent_available = True
except ImportError:
    _gevent_available = False

# Handle current green mode

try:
    _CURRENT_GREEN_MODE = getattr(
        GreenMode, os.environ["PYTANGO_GREEN_MODE"].capitalize()
    )
except Exception:
    _CURRENT_GREEN_MODE = GreenMode.Synchronous


def set_green_mode(green_mode=None):
    """Sets the global default PyTango green mode.

    Advice: Use only in your final application. Don't use this in a python
    library in order not to interfere with the beavior of other libraries
    and/or application where your library is being.

    :param green_mode: the new global default PyTango green mode
    :type green_mode: GreenMode
    """
    global _CURRENT_GREEN_MODE
    # Make sure the green mode is available
    get_executor(green_mode)
    # Set the green mode
    _CURRENT_GREEN_MODE = green_mode


def get_green_mode():
    """Returns the current global default PyTango green mode.

    :returns: the current global default PyTango green mode
    :rtype: GreenMode
    """
    return _CURRENT_GREEN_MODE


# Abstract executor class


class AbstractExecutor:
    asynchronous = NotImplemented
    default_wait = NotImplemented

    def __init__(self):
        self.ident = get_ident(), os.getpid()

    def get_ident(self):
        return self.ident

    def in_executor_context(self):
        return self.ident == (get_ident(), os.getpid())

    def delegate(self, fn, *args, **kwargs):
        """Delegate an operation and return an accessor."""
        if not self.asynchronous:
            raise ValueError("Not supported in synchronous mode")
        raise NotImplementedError

    def access(self, accessor, timeout=None):
        """Return a result from an accessor."""
        if not self.asynchronous:
            raise ValueError("Not supported in synchronous mode")
        raise NotImplementedError

    def submit(self, fn, *args, **kwargs):
        """Submit an operation"""
        if not self.asynchronous:
            return fn(*args, **kwargs)
        raise NotImplementedError

    def execute(self, fn, *args, **kwargs):
        """Execute an operation and return the result."""
        if not self.asynchronous:
            return fn(*args, **kwargs)
        raise NotImplementedError

    def run(self, fn, args=(), kwargs={}, wait=None, timeout=None):
        if wait is None:
            wait = self.default_wait
        # Wait and timeout are not supported in synchronous mode
        if not self.asynchronous and (not wait or timeout):
            raise ValueError("Not supported in synchronous mode")
        # Synchronous (no delegation)
        if not self.asynchronous or not self.in_executor_context():
            return fn(*args, **kwargs)
        # Asynchronous delegation
        accessor = self.delegate(fn, *args, **kwargs)
        if not wait:
            return accessor
        return self.access(accessor, timeout=timeout)


class SynchronousExecutor(AbstractExecutor):
    asynchronous = False
    default_wait = True


# Default synchronous executor


def get_synchronous_executor():
    return _SYNCHRONOUS_EXECUTOR


_SYNCHRONOUS_EXECUTOR = SynchronousExecutor()


# Getters


def get_object_green_mode(obj):
    if hasattr(obj, "get_green_mode"):
        return obj.get_green_mode()
    return get_green_mode()


def get_executor(green_mode=None):
    if green_mode is None:
        green_mode = get_green_mode()
    # Valid green modes
    if green_mode == GreenMode.Synchronous:
        return get_synchronous_executor()
    if green_mode == GreenMode.Gevent:
        from tango import gevent_executor

        return gevent_executor.get_global_executor()
    if green_mode == GreenMode.Futures:
        from tango import futures_executor

        return futures_executor.get_global_executor()
    if green_mode == GreenMode.Asyncio:
        from tango import asyncio_executor

        return asyncio_executor.get_global_executor()
    # Invalid green mode
    raise TypeError("Not a valid green mode")


def switch_existing_global_executors_to_thread():
    """
    checks which global executor existing, and if they are belong to the caller thread
    if not - creates a new executor, linked to thread, and set it as global
    """
    from tango import asyncio_executor
    from tango import futures_executor

    if _gevent_available:
        from tango import gevent_executor
    else:
        gevent_executor = None

    for executor in [asyncio_executor, futures_executor, gevent_executor]:
        if executor:
            executor._switch_global_executor_to_thread()


def get_object_executor(obj, green_mode=None):
    """Returns the proper executor for the given object.

    If the object has *_executors* and *_green_mode* members it returns
    the submit callable for the executor corresponding to the green_mode.
    Otherwise it returns the global executor for the given green_mode.

    Note: *None* is a valid object.

    :returns: submit callable"""
    # Get green mode
    if green_mode is None:
        green_mode = get_object_green_mode(obj)
    # Get executor
    executor = None
    if hasattr(obj, "_executors"):
        executor = obj._executors.get(green_mode, None)
    if executor is None:
        executor = get_executor(green_mode)
    # Get submitter
    return executor


# Green modifiers


def green(fn=None, consume_green_mode=True, update_signature_and_docstring=False):
    """Make a function green. Can be used as a decorator."""

    def decorator(fn):
        @wraps(fn)
        def greener(obj, *args, **kwargs):
            args = (obj,) + args
            wait = kwargs.pop("wait", None)
            timeout = kwargs.pop("timeout", None)
            access = kwargs.pop if consume_green_mode else kwargs.get
            green_mode = access("green_mode", None)
            executor = get_object_executor(obj, green_mode)
            return executor.run(fn, args, kwargs, wait=wait, timeout=timeout)

        sig = inspect.signature(fn)

        if update_signature_and_docstring:

            # Build green parameters
            green_mode_param = inspect.Parameter(
                "green_mode",
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=None,
                annotation=None,
            )

            wait_param = inspect.Parameter(
                "wait",
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=None,
                annotation=None,
            )

            timeout_param = inspect.Parameter(
                "timeout",
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=None,
                annotation=None,
            )

            # Append it to the existing parameters
            old_params = list(sig.parameters.values())
            add_kwargs = False
            if old_params[-1].kind == inspect.Parameter.VAR_KEYWORD:
                old_params = old_params[:-1]
                add_kwargs = True

            if "green_mode" in [param.name for param in old_params]:
                new_params = old_params + [wait_param, timeout_param]
            else:
                new_params = old_params + [green_mode_param, wait_param, timeout_param]

            if add_kwargs:
                new_params += [
                    inspect.Parameter("kwargs", kind=inspect.Parameter.VAR_KEYWORD)
                ]

            new_sig = sig.replace(parameters=new_params)

            if greener.__doc__ is not None:
                fill_green_doc(greener)

            greener.__signature__ = new_sig

        else:
            greener.__signature__ = sig

        return greener

    if fn is None:
        return decorator
    return decorator(fn)


def green_callback(fn, obj=None, green_mode=None):
    """Return a green verion of the given callback."""
    executor = get_object_executor(obj, green_mode)

    @wraps(fn)
    def greener(*args, **kwargs):
        return executor.submit(_forcefully_traced_method(fn), *args, **kwargs)

    return greener


__GREEN_KWARGS__ = "green_mode=None, wait=True, timeout=None"
__GREEN_KWARGS_DESCRIPTION__ = """
:param green_mode: Defaults to the current tango GreenMode. Refer to :meth:`~tango.DeviceProxy.get_green_mode` and :meth:`~tango.DeviceProxy.set_green_mode` for more details.
:type green_mode: :obj:`tango.GreenMode`, optional

:param wait: Specifies whether to wait for the result. If `green_mode` is *Synchronous*, this parameter is ignored as the operation always waits for the result. This parameter is also ignored when `green_mode` is Synchronous.
:type wait: bool, optional

:param timeout: The number of seconds to wait for the result. If set to `None`, there is no limit on the wait time. This parameter is ignored when `green_mode` is Synchronous or when `wait` is False.
:type timeout: float, optional
"""
__GREEN_RAISES__ = """
:throws: :obj:`TimeoutError`: (green_mode == Futures) If the future didn't finish executing before the given timeout.
:throws: :obj:`Timeout`: (green_mode == Gevent) If the async result didn't finish executing before the given timeout.
"""


def fill_green_doc(method):
    """
    Replace the __GREEN_KWARGS__ __GREEN_KWARGS_DESCRIPTION__ placeholders in `doc`
    preserving the placeholder’s indentation.
    """

    dedented = textwrap.dedent(method.__doc__)
    dedented = dedented.replace("__GREEN_KWARGS__", __GREEN_KWARGS__)

    m = re.search(
        r"^(?P<indent>[ \t]*)__GREEN_KWARGS_DESCRIPTION__", dedented, flags=re.MULTILINE
    )
    if not m:
        return

    indent = m.group("indent")

    indented_desc = textwrap.indent(__GREEN_KWARGS_DESCRIPTION__.strip("\n"), indent)
    indented_raises = textwrap.indent(__GREEN_RAISES__.strip("\n"), indent)

    dedented = re.sub(
        r"^[ \t]*__GREEN_KWARGS_DESCRIPTION__",
        indented_desc,
        dedented,
        flags=re.MULTILINE,
    )

    method.__doc__ = re.sub(
        r"^[ \t]*__GREEN_RAISES__", indented_raises, dedented, flags=re.MULTILINE
    )



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Power supply device that connects to a "hardware" simulator
provided by ps-simulator.py.  Launch the ps-simulator.py script
before starting this device server.

Uses the gevent green mode.
"""

from gevent import sleep
from gevent.socket import create_connection

from tango.server import Device, attribute, command, device_property


def connect(host, port):
    sock = create_connection((host, port))
    return sock.makefile("rwb", newline=b"\n", buffering=0)


def write_readline(conn, msg):
    conn.write(msg)
    return conn.readline()


class PowerSupply(Device):
    host = device_property(str, default_value="localhost")
    port = device_property(int, default_value=45000)

    def init_device(self):
        super().init_device()
        self.conn = connect(self.host, self.port)

    @attribute(dtype=float)
    def voltage(self):
        return float(write_readline(self.conn, b"VOL?\n"))

    @command
    def calibrate(self):
        write_readline(self.conn, b"CALIB 1\n")
        while int(write_readline(self.conn, b"stat?\n")):
            sleep(0.1)


if __name__ == "__main__":
    from tango import GreenMode

    PowerSupply.run_server(green_mode=GreenMode.Gevent)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import time
from PyTango.server import Device, attribute, command


class Clock(Device):
    @attribute
    def time(self):
        return time.time()

    @command(dtype_in=str, dtype_out=str)
    def strftime(self, format):
        return time.strftime(format)


if __name__ == "__main__":
    Clock.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""Client tests that run against the standard TangoTest device.

Due to a TANGO 9 issue (#821), the device is run without any database.
Note that this means that various features won't work:

 * No device configuration via properties.
 * No event generated by the server.
 * No memorized attributes.
 * No device attribute configuration via the database.

So don't even try to test anything of the above as it will not work
and is even likely to crash the device (!)

"""

import asyncio
import gc
import multiprocessing
import sys
import time
import weakref

import pytest

from tango import (
    AttributeInfo,
    AttributeInfoEx,
    AttributeInfoList,
    AttributeInfoListEx,
    AttrWriteType,
    DeviceInfo,
    DeviceProxy,
    DevFailed,
    EventType,
    ExtractAs,
    GreenMode,
    Group,
    PyTangoUserWarning,
    constants,
    EventSubMode,
)
from tango.constants import AllAttr
from tango.server import Device, attribute, command
from tango.test_utils import (
    DeviceTestContext,
    assert_close,
    bytes_devstring,
    convert_to_type,
    str_devstring,
)
from tango.utils import (
    EventCallback,
    is_str_type,
    is_int_type,
    is_float_type,
    is_bool_type,
)

ATTRIBUTES = [
    "State",
    "Status",
    "ampli",
    "boolean_image",
    "boolean_image_ro",
    "boolean_scalar",
    "boolean_spectrum",
    "boolean_spectrum_ro",
    "double_image",
    "double_image_ro",
    "double_scalar",
    "double_scalar_rww",
    "double_scalar_w",
    "double_spectrum",
    "double_spectrum_ro",
    "echo_mode",
    "enum_image",
    "enum_image_ro",
    "enum_scalar",
    "enum_scalar_ro",
    "enum_spectrum",
    "enum_spectrum_ro",
    "float_image",
    "float_image_ro",
    "float_scalar",
    "float_spectrum",
    "float_spectrum_ro",
    "freq",
    "long64_image_ro",
    "long64_scalar",
    "long64_spectrum_ro",
    "long_image",
    "long_image_ro",
    "long_scalar",
    "long_scalar_rww",
    "long_scalar_w",
    "long_spectrum",
    "long_spectrum_ro",
    "no_value",
    "short_image",
    "short_image_ro",
    "short_scalar",
    "short_scalar_ro",
    "short_scalar_rww",
    "short_scalar_w",
    "short_spectrum",
    "short_spectrum_ro",
    "string_image",
    "string_image_ro",
    "string_scalar",
    "string_spectrum",
    "string_spectrum_ro",
    "throw_exception",
    "uchar_image",
    "uchar_image_ro",
    "uchar_scalar",
    "uchar_spectrum",
    "uchar_spectrum_ro",
    "ulong64_image_ro",
    "ulong64_scalar",
    "ulong64_spectrum_ro",
    "ulong_image_ro",
    "ulong_scalar",
    "ulong_spectrum_ro",
    "ushort_image",
    "ushort_image_ro",
    "ushort_scalar",
    "ushort_spectrum",
    "ushort_spectrum_ro",
    "wave",
]

WRITABLE_SCALAR_ATTRIBUTES = [
    a for a in ATTRIBUTES if "scalar" in a and a.split("_")[-1] not in ("ro", "rww")
]
WRITABLE_SPECTRUM_ATTRIBUTES = [
    a for a in ATTRIBUTES if "spectrum" in a and a.split("_")[-1] not in ("ro", "rww")
]

TEST_DOUBLE_ATTRIBUTES = (
    ("double_scalar_w", -28.2),
    ("double_spectrum", [-28.2, 23.4]),
    ("double_image", [[-28.2, 23.4], [-4.9, 6.5]]),
)

# Helpers


def ping_device(proxy):
    if proxy.get_green_mode() == GreenMode.Asyncio:
        asyncio.get_event_loop().run_until_complete(proxy.ping())
    else:
        proxy.ping()


# Fixtures


@pytest.fixture(params=ATTRIBUTES)
def attributes(request):
    return request.param


@pytest.fixture(
    params=[a for a in ATTRIBUTES if a not in ("no_value", "throw_exception")]
)
def readable_attribute(request):
    return request.param


@pytest.fixture(params=WRITABLE_SCALAR_ATTRIBUTES)
def writable_scalar_attribute(request):
    return request.param


@pytest.fixture(params=WRITABLE_SPECTRUM_ATTRIBUTES)
def writable_spectrum_attribute(request):
    return request.param


@pytest.fixture
def simple_device_fqdn():
    class TestDevice(Device):
        pass

    context = DeviceTestContext(TestDevice, host="127.0.0.1")
    context.start()
    yield context.get_device_access()
    context.stop()


# Tests


def test_ping(tango_test):
    duration = tango_test.ping(wait=True)
    assert isinstance(duration, int)


def test_info(tango_test):
    info = tango_test.info()
    assert isinstance(info, DeviceInfo)
    assert info.dev_class == "TangoTest"
    info_dict = info.version_info
    assert isinstance(info_dict, dict)
    # version_info only populated from IDL v6
    if info.server_version >= 6:
        assert len(info_dict) > 0
    else:
        assert len(info_dict) == 0


def test_read_attribute(tango_test, readable_attribute):
    "Check that readable attributes can be read"
    # For read-only string spectrum and read-only string image types,
    # the following error is very likely to be raised:
    # -> MARSHAL CORBA system exception: MARSHAL_PassEndOfMessage
    # An explicit sleep fixes the problem, but it's annoying to maintain

    if readable_attribute in ["string_image_ro", "string_spectrum_ro"]:
        pytest.xfail()
    tango_test.read_attribute(readable_attribute, wait=True)


def test_read_write_attribute_with_green_modes(tango_test_with_green_modes):
    """
    Check read/write to attributes one at a time with all green modes
    """
    for attr_name, write_value in TEST_DOUBLE_ATTRIBUTES:
        tango_test_with_green_modes.write_attribute(attr_name, write_value, wait=True)
        read_attr = tango_test_with_green_modes.read_attribute(attr_name, wait=True)
        assert_close(read_attr.value, write_value)
        assert_close(read_attr.w_value, write_value)


def test_read_write_attribute_with_attr_info(tango_test):
    for attr_name, write_value in TEST_DOUBLE_ATTRIBUTES:
        attr_info = tango_test.get_attribute_config(attr_name, wait=True)
        tango_test.write_attribute(attr_info, write_value, wait=True)
        read_attr = tango_test.read_attribute(attr_name, wait=True)
        assert_close(read_attr.value, write_value)
        assert_close(read_attr.w_value, write_value)


def test_read_write_attributes_with_green_modes(tango_test_with_green_modes):
    """
    Check read/write to multiple attributes with all green modes
    """
    tango_test_with_green_modes.write_attributes(TEST_DOUBLE_ATTRIBUTES, wait=True)
    attr_names = [name for name, _ in TEST_DOUBLE_ATTRIBUTES]
    read_attrs = tango_test_with_green_modes.read_attributes(attr_names, wait=True)
    for (attr_name, write_value), read_attr in zip(TEST_DOUBLE_ATTRIBUTES, read_attrs):
        assert read_attr.name == attr_name
        assert_close(read_attr.value, write_value)
        assert_close(read_attr.w_value, write_value)


def test_read_write_attributes_with_attr_info(tango_test):
    attr_names = [name for name, _ in TEST_DOUBLE_ATTRIBUTES]
    attr_infos = tango_test.get_attribute_config_ex(attr_names)

    value_to_write = [
        (attr_info, value)
        for (attr_info, (_, value)) in zip(attr_infos, TEST_DOUBLE_ATTRIBUTES)
    ]

    tango_test.write_attributes(value_to_write, wait=True)
    read_attrs = tango_test.read_attributes(attr_names, wait=True)
    for (attr_name, write_value), read_attr in zip(TEST_DOUBLE_ATTRIBUTES, read_attrs):
        assert read_attr.name == attr_name
        assert_close(read_attr.value, write_value)
        assert_close(read_attr.w_value, write_value)


@pytest.mark.parametrize("sequence_type", ["tuple", "list"])
def test_read_write_attributes_different_sequences(tango_test, sequence_type):
    """
    Check read/write to multiple attributes with different sequences
    """
    attr_names_values = ((name, value) for (name, value) in TEST_DOUBLE_ATTRIBUTES)
    attr_names = (name for name, _ in TEST_DOUBLE_ATTRIBUTES)

    if sequence_type == "tuple":
        attr_names_values = tuple(attr_names_values)
        attr_names = tuple(attr_names)
    elif sequence_type == "list":
        attr_names_values = list(attr_names_values)
        attr_names = list(attr_names)

    tango_test.write_attributes(attr_names_values, wait=True)
    read_attrs = tango_test.read_attributes(attr_names, wait=True)
    for (attr_name, write_value), read_attr in zip(TEST_DOUBLE_ATTRIBUTES, read_attrs):
        assert read_attr.name == attr_name
        assert_close(read_attr.value, write_value)
        assert_close(read_attr.w_value, write_value)


def test_write_read_attributes_with_green_modes(tango_test_with_green_modes):
    """
    Check write and read multiple attributes in a single call with all green modes
    """
    attr_names = [name for name, _ in TEST_DOUBLE_ATTRIBUTES]
    read_attrs = tango_test_with_green_modes.write_read_attributes(
        TEST_DOUBLE_ATTRIBUTES, attr_names, wait=True
    )
    for (attr_name, write_value), read_attr in zip(TEST_DOUBLE_ATTRIBUTES, read_attrs):
        assert read_attr.name == attr_name
        assert_close(read_attr.value, write_value)
        assert_close(read_attr.w_value, write_value)


@pytest.mark.asyncio
async def test_high_level_api_for_asyncio(tango_test):
    tango_test.set_green_mode(GreenMode.Asyncio)
    _ = await tango_test.long_scalar
    _ = await getattr(tango_test, "long_scalar")
    _ = await tango_test["long_scalar"]


def test_write_scalar_attribute(tango_test, writable_scalar_attribute):
    "Check that writable scalar attributes can be written"
    attr_name = writable_scalar_attribute
    config = tango_test.get_attribute_config(attr_name, wait=True)
    if is_bool_type(config.data_type):
        tango_test.write_attribute(attr_name, True, wait=True)
    elif is_int_type(config.data_type):
        tango_test.write_attribute(attr_name, 76, wait=True)
    elif is_float_type(config.data_type):
        tango_test.write_attribute(attr_name, -28.2, wait=True)
    elif is_str_type(config.data_type):
        tango_test.write_attribute(attr_name, "hello", wait=True)
    else:
        pytest.xfail("Not currently testing this type")


def test_write_read_spectrum_attribute(
    tango_test, writable_spectrum_attribute, extract_as
):
    "Check that writable spectrum attributes can be written and read"
    requested_type, expected_type = extract_as
    attr_name = writable_spectrum_attribute
    config = tango_test.get_attribute_config(attr_name, wait=True)
    if is_bool_type(config.data_type):
        write_values = [True, False]
    elif is_int_type(config.data_type):
        write_values = [76, 77]
    elif is_float_type(config.data_type):
        if requested_type == ExtractAs.String:
            # it is hard to find a proper float values, which can be converted to str,
            # most of them case UnicodeDecodeError, so we use the most simple one
            write_values = [0, 0]
        else:
            write_values = [-28.2, 44.3]
    elif is_str_type(config.data_type):
        if requested_type == ExtractAs.Numpy:
            expected_type = tuple
        if requested_type in [ExtractAs.ByteArray, ExtractAs.Bytes, ExtractAs.String]:
            pytest.xfail(
                "Conversion from (str,) to ByteArray, Bytes and String not supported. May be fixed in future"
            )
        write_values = ["hello", "hola"]
    else:
        pytest.xfail("Not currently testing this type")

    tango_test.write_attribute(attr_name, write_values, wait=True)
    read_attr = tango_test.read_attribute(attr_name, extract_as=requested_type)

    assert isinstance(read_attr.value, expected_type)
    assert_close(
        read_attr.value, convert_to_type(write_values, config.data_type, expected_type)
    )

    assert isinstance(read_attr.w_value, expected_type)
    assert_close(
        read_attr.w_value,
        convert_to_type(write_values, config.data_type, expected_type),
    )


def test_write_read_empty_spectrum_attribute(tango_test, writable_spectrum_attribute):
    "Check that writing empty list to spectrum attribute reads back as None."
    attr_name = writable_spectrum_attribute
    config = tango_test.get_attribute_config(attr_name, wait=True)
    if is_str_type(config.data_type):
        pytest.xfail(
            "Conversion from (str,) to numpy not supported. Probably, may be fixed in future"
        )

    tango_test.write_attribute(attr_name, [], wait=True)
    read_attr = tango_test.read_attribute(attr_name, wait=True)
    assert not len(read_attr.value)


def test_write_read_string_attribute(tango_test):
    attr_name = "string_scalar"
    bytes_big = 100000 * b"big data "
    str_big = bytes_big.decode("latin-1")

    values = [
        b"",
        "",
        "Hello, World!",
        b"Hello, World!",
        bytes_devstring,
        str_devstring,
        bytes_big,
        str_big,
    ]

    expected_values = [
        "",
        "",
        "Hello, World!",
        "Hello, World!",
        str_devstring,
        str_devstring,
        str_big,
        str_big,
    ]

    for value, expected_value in zip(values, expected_values):
        tango_test.write_attribute(attr_name, value, wait=True)
        result = tango_test.read_attribute(attr_name, wait=True)
        assert result.value == expected_value

    attr_name = "string_spectrum"
    for value, expected_value in zip(values, expected_values):
        tango_test.write_attribute(attr_name, ["", value, ""], wait=True)
        result = tango_test.read_attribute(attr_name, wait=True)
        assert result.value[1] == expected_value

    attr_name = "string_image"
    for value, expected_value in zip(values, expected_values):
        tango_test.write_attribute(attr_name, [[value], [value]], wait=True)
        result = tango_test.read_attribute(attr_name, wait=True)
        assert result.value == ((expected_value,), (expected_value,))


def test_set_non_existent_attribute_raises_by_default(tango_test):
    with pytest.raises(AttributeError, match="some_invalid_name"):
        tango_test.some_invalid_name = "123"


def test_set_non_existent_attribute_allowed_if_dynamic_interface_unfrozen(tango_test):
    with pytest.warns(PyTangoUserWarning):
        tango_test.unfreeze_dynamic_interface()
    tango_test.some_invalid_name = "123"
    assert tango_test.some_invalid_name == "123"


def test_dynamic_interface_can_be_toggled(tango_test):
    with pytest.warns(PyTangoUserWarning):
        tango_test.unfreeze_dynamic_interface()
    tango_test.some_invalid_name = "456"
    assert tango_test.some_invalid_name == "456"
    tango_test.freeze_dynamic_interface()
    with pytest.raises(AttributeError, match="another_invalid_name"):
        tango_test.another_invalid_name = "123"


def test_dynamic_interface_flag_can_be_read(tango_test):
    with pytest.warns(PyTangoUserWarning):
        tango_test.unfreeze_dynamic_interface()
    assert not tango_test.is_dynamic_interface_frozen()
    tango_test.freeze_dynamic_interface()
    assert tango_test.is_dynamic_interface_frozen()


def test_dynamic_interface_only_applies_to_device_proxy_instance(tango_test):
    other_proxy = DeviceProxy(tango_test.adm_name())
    with pytest.warns(PyTangoUserWarning):
        tango_test.unfreeze_dynamic_interface()
    other_proxy.freeze_dynamic_interface()
    assert not tango_test.is_dynamic_interface_frozen()
    assert other_proxy.is_dynamic_interface_frozen()


def test_dynamic_interface_unfreeze_generates_a_user_warning(tango_test):
    with pytest.warns(PyTangoUserWarning):
        tango_test.unfreeze_dynamic_interface()


def test_get_attribute_config(tango_test):
    all_conf = tango_test.get_attribute_config(ATTRIBUTES)
    assert len(all_conf) == len(ATTRIBUTES)
    assert set([conf.name for conf in all_conf]) == set(ATTRIBUTES)

    for attr in ATTRIBUTES:
        tango_test.get_attribute_config(attr)


def test_get_attribute_config_ex():
    class TestDevice(Device):

        @attribute(dtype=int, unit="mA")
        def attr_config_int(self):
            return 1

        @attribute(dtype=bool)
        def attr_config_bool(self):
            return True

    def assert_attr_config_ok(dev_proxy):
        # testing that call to get_attribute_config_ex for all types of
        # input arguments gives same result and doesn't raise an exception
        ac1 = dev_proxy.get_attribute_config_ex("attr_config_int")
        ac2 = dev_proxy.get_attribute_config_ex(["attr_config_int"])
        assert repr(ac1) == repr(ac2)

    def assert_multiple_attrs_config_ok(dev_proxy):
        # testing that querying multiple attributes gives same result and
        # doesn't raise an exception
        ac1 = dev_proxy.get_attribute_config_ex("attr_config_int")
        ac2 = dev_proxy.get_attribute_config_ex("attr_config_bool")
        acs = dev_proxy.get_attribute_config_ex(("attr_config_int", "attr_config_bool"))
        acs_2 = dev_proxy.get_attribute_config_ex(
            ["attr_config_int", "attr_config_bool"]
        )

        acs_all = dev_proxy.get_attribute_config_ex(AllAttr)

        assert repr(ac1[0]) == repr(acs[0]) == repr(acs_2[0]) == repr(
            acs_all[0]
        ) and repr(ac2[0]) == repr(acs[1]) == repr(acs_2[1]) == repr(acs_all[1])

    with DeviceTestContext(TestDevice) as proxy:
        assert_attr_config_ok(proxy)
        assert_multiple_attrs_config_ok(proxy)


def test_set_attribute_config():
    class TestDevice(Device):

        @attribute(dtype=int, unit="mA")
        def attr_config(self):
            return 1

    with DeviceTestContext(TestDevice) as proxy:
        attr_info = proxy.get_attribute_config("attr_config")
        attr_info.unit = "testing"
        proxy.set_attribute_config(attr_info)

        updated_attr_info = proxy.get_attribute_config("attr_config")
        assert updated_attr_info.unit == "testing"


def test_set_attribute_config_signatures(tango_test):
    settable_attrs = ATTRIBUTES.copy()
    settable_attrs.remove("State")
    settable_attrs.remove("Status")
    all_conf = tango_test.get_attribute_config(settable_attrs)
    all_conf_ex = tango_test.get_attribute_config_ex(settable_attrs)

    tango_test.set_attribute_config(all_conf)
    tango_test.set_attribute_config(all_conf[0])
    tango_test.set_attribute_config(all_conf_ex)
    tango_test.set_attribute_config(all_conf_ex[0])
    tango_test.set_attribute_config([all_conf[0], all_conf[1]])
    tango_test.set_attribute_config([all_conf_ex[0], all_conf_ex[1]])

    with pytest.raises(TypeError, match=r"AttributeInfo.*AttributeInfoEx"):
        tango_test.set_attribute_config("invalid_item")
    with pytest.raises(TypeError, match=r"AttributeInfo.*AttributeInfoEx"):
        tango_test.set_attribute_config(["invalid", "list", "of", "items"])


def test_attribute_list_query(tango_test):
    attrs = tango_test.attribute_list_query()
    assert isinstance(attrs, AttributeInfoList)
    assert all(isinstance(a, AttributeInfo) for a in attrs)
    assert {a.name for a in attrs} == set(ATTRIBUTES)


def test_attribute_list_query_ex(tango_test):
    attrs = tango_test.attribute_list_query_ex()
    assert isinstance(attrs, AttributeInfoListEx)
    assert all(isinstance(a, AttributeInfoEx) for a in attrs)
    assert {a.name for a in attrs} == set(ATTRIBUTES)


def test_device_proxy_dir_method(tango_test):
    lst = dir(tango_test)
    attrs = tango_test.get_attribute_list()
    cmds = tango_test.get_command_list()
    methods = dir(type(tango_test))
    internals = tango_test.__dict__.keys()
    # Check attributes
    assert set(attrs) < set(lst)
    assert set(map(str.lower, attrs)) < set(lst)
    # Check commands
    assert set(cmds) < set(lst)
    assert set(map(str.lower, cmds)) < set(lst)
    # Check internals
    assert set(methods) <= set(lst)
    # Check internals
    assert set(internals) <= set(lst)


def test_device_polling_command(tango_test):
    dct = {"SwitchStates": 1000, "DevVoid": 10000, "DumpExecutionState": 5000}

    for cmd, period in dct.items():
        tango_test.poll_command(cmd, period)

    ans = tango_test.polling_status()
    for info in ans:
        lines = info.split("\n")
        cmd = lines[0].split("= ")[1]
        period = int(lines[1].split("= ")[1])
        assert dct[cmd] == period


def test_device_polling_attribute(tango_test):
    dct = {"boolean_scalar": 1000, "double_scalar": 10000, "long_scalar": 5000}

    for attr, poll_period in dct.items():
        tango_test.poll_attribute(attr, poll_period)

    ans = tango_test.polling_status()
    for x in ans:
        lines = x.split("\n")
        attr = lines[0].split("= ")[1]
        poll_period = int(lines[1].split("= ")[1])
        assert dct[attr] == poll_period


def test_command_string(tango_test):
    cmd_name = "DevString"
    bytes_big = 100000 * b"big data "
    str_big = bytes_big.decode("latin-1")

    values = [
        b"",
        "",
        "Hello, World!",
        b"Hello, World!",
        bytes_devstring,
        str_devstring,
        bytes_big,
        str_big,
    ]

    expected_values = [
        "",
        "",
        "Hello, World!",
        "Hello, World!",
        str_devstring,
        str_devstring,
        str_big,
        str_big,
    ]

    for value, expected_value in zip(values, expected_values):
        result = tango_test.command_inout(cmd_name, value, wait=True)
        assert result == expected_value

    cmd_name = "DevVarStringArray"
    for value, expected_value in zip(values, expected_values):
        result = tango_test.command_inout(cmd_name, [value, value], wait=True)
        assert result == [expected_value, expected_value]

    cmd_name = "DevVarLongStringArray"
    for value, expected_value in zip(values, expected_values):
        result = tango_test.command_inout(
            cmd_name, [[-10, 200], [value, value]], wait=True
        )
        assert len(result) == 2
        assert_close(result[0], [-10, 200])
        assert_close(result[1], [expected_value, expected_value])


def test_command_raises_type_error_for_bad_input(tango_test):
    expected_message = "Invalid input argument for command"
    with pytest.raises(TypeError, match=expected_message):
        tango_test.command_inout("DevDouble", "123.0")
    with pytest.raises(TypeError, match=expected_message):
        tango_test.command_inout("DevLong64", [1, 2, 3])
    with pytest.raises(TypeError, match=expected_message):
        tango_test.command_inout("DevString", [{"invalid": "type"}, 123.0])
    with pytest.raises(TypeError, match=expected_message):
        tango_test.command_inout("DevVarStringArray", [1, 2, 3])
    with pytest.raises(TypeError, match=expected_message):
        tango_test.command_inout("DevVarLong64Array", ["1", 2.0, {3}])


def test_repr_uses_info(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    assert repr(proxy) == "TestDevice(test/nodb/testdevice)"


def test_repr_default_if_info_unavailable(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)

    def bad_info(self):
        raise RuntimeError("Break info for test")

    proxy.__class__.info = bad_info

    assert repr(proxy) == "Device(test/nodb/testdevice)"


def test_multiple_repr_calls_only_call_info_once(
    green_mode_device_proxy, simple_device_fqdn
):
    proxy = green_mode_device_proxy(simple_device_fqdn)

    def mock_info(self):
        self.info_call_count += 1
        return self.info_orig()

    proxy.__class__.info_orig = proxy.info
    proxy.__class__.info = mock_info
    proxy.__dict__["info_call_count"] = 0

    repr(proxy)
    assert proxy.info_call_count == 1
    repr(proxy)
    assert proxy.info_call_count == 1


def test_no_memory_leak_for_repr(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    ping_device(proxy)
    weak_ref = weakref.ref(proxy)

    repr(proxy)

    # clear strong reference and check if object can be garbage collected
    del proxy
    assert_object_released_after_gc(weak_ref)


def test_no_memory_leak_for_str(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    ping_device(proxy)
    weak_ref = weakref.ref(proxy)

    str(proxy)

    # clear strong reference and check if object can be garbage collected
    del proxy
    assert_object_released_after_gc(weak_ref)


def test_no_cyclic_ref_for_proxy(green_mode_device_proxy, simple_device_fqdn):
    proxy = green_mode_device_proxy(simple_device_fqdn)
    ping_device(proxy)
    weak_ref = weakref.ref(proxy)

    # clear strong reference and check if object is immediately released
    del proxy
    assert_object_released_without_gc(weak_ref)


def group_client_lifecycle(device_trl):
    group = Group("test")
    group_weak_ref = weakref.ref(group)
    group.add(device_trl)
    del group
    assert_object_released_without_gc(group_weak_ref)


def device_proxy_lifecycle(device_trl):
    proxy = DeviceProxy(device_trl)
    proxy_weak_ref = weakref.ref(proxy)
    del proxy
    assert_object_released_without_gc(proxy_weak_ref)


@pytest.mark.extra_src_test
@pytest.mark.parametrize("subject", [group_client_lifecycle, device_proxy_lifecycle])
def test_client_destructor_does_not_deadlock(
    tango_test_process_device_trl_with_function_scope, subject
):
    proc, device_trl = tango_test_process_device_trl_with_function_scope

    # Create client to be run in a subprocess.
    # We want this client to wait for a "not connected event"
    # while rapidly creating and destroying objects (using "subject" function)
    # to try to induce a race condition.
    # The "not connected event" is about 2 seconds after the event heartbeat period
    # of 10 seconds - we add some margin either side.
    # We use some Event objects for additional synchronisation
    heartbeat_margin = 0.5
    delay_before_heartbeat = constants.EVENT_HEARTBEAT_PERIOD - heartbeat_margin
    max_time_until_not_connected_event = delay_before_heartbeat + 3.0
    process_startup_timeout = 1.0
    process_exit_timeout = 1.0
    proxy_created = multiprocessing.Event()
    process_terminated = multiprocessing.Event()
    client = multiprocessing.Process(
        target=continuously_call_subject_during_not_connected_event,
        args=(
            device_trl,
            proxy_created,
            process_terminated,
            delay_before_heartbeat,
            max_time_until_not_connected_event,
            subject,
        ),
    )
    # start client, and wait for it to create a DeviceProxy
    client.start()
    assert proxy_created.wait(timeout=process_startup_timeout)

    # now that client has created DeviceProxy, we terminate the device server process
    # and inform the client
    proc.terminate()
    process_terminated.set()

    # finally, wait for client.
    # - if deadlocked, the exitcode will be None.
    # - if it has some other problem, the exitcode will be non-zero.
    client.join(timeout=max_time_until_not_connected_event + process_exit_timeout)
    assert client.exitcode == 0


def continuously_call_subject_during_not_connected_event(
    device_trl: str,
    proxy_created: multiprocessing.Event,
    process_terminated: multiprocessing.Event,
    delay_before_heartbeat: float,
    max_duration: float,
    subject: callable,
):
    # stateless subscription
    cb = EventCallback()
    proxy = DeviceProxy(device_trl)
    proxy_created.set()
    eid = proxy.subscribe_event(
        "boolean_scalar", EventType.CHANGE_EVENT, cb, sub_mode=EventSubMode.Stateless
    )
    assert eid > 0

    # wait for Tango device server to be terminated by parent
    terminate_timeout = 1.0
    assert process_terminated.wait(timeout=terminate_timeout)

    # wait for a "not connected event" while rapidly calling the
    # subject function (which will create and destroy objects)
    # to try to induce the race condition
    start_time = time.time()
    time.sleep(delay_before_heartbeat)
    while time.time() - start_time < max_duration:
        subject(device_trl)


def assert_object_released_after_gc(weak_ref):
    gc.collect()
    assert weak_ref() is None


def assert_object_released_without_gc(weak_ref):
    try:
        assert weak_ref() is None
    except AssertionError:
        if weak_ref().get_green_mode() == GreenMode.Asyncio:
            pytest.xfail("Sometimes fails with a concurrent.Future ref cycle")
        else:
            raise


@pytest.fixture
def uninitialized_dev_proxy():
    """
    This could happen if there was an exception in the DeviceProxy __init__ method,
    and the user is running through pytest.  pytest will have a reference to the frame
    and try to print out all the objects from the failed test.
    """
    proxy_instance = None
    try:
        DeviceProxy("not/existing/device")
    except DevFailed:
        traceback = sys.exc_info()[2]
        for v in traceback.tb_next.tb_frame.f_locals.values():
            if isinstance(v, DeviceProxy):
                proxy_instance = v

    assert proxy_instance is not None
    yield proxy_instance


def test_pytest_report_on_failed_device_proxy_does_not_crash(uninitialized_dev_proxy):
    safe_repr = repr(uninitialized_dev_proxy)
    assert "DeviceProxy" in safe_repr
    assert "Unknown" in safe_repr

    safe_str = str(uninitialized_dev_proxy)
    assert "DeviceProxy" in safe_str
    assert "Unknown" in safe_str


def test_dunder_methods_on_failed_device_proxy_do_not_crash(uninitialized_dev_proxy):
    with pytest.raises(RuntimeError, match="not fully initialized"):
        dir(uninitialized_dev_proxy)

    with pytest.raises(RuntimeError, match="not fully initialized"):
        getattr(uninitialized_dev_proxy, "dummy")

    with pytest.raises(RuntimeError, match="not fully initialized"):
        setattr(uninitialized_dev_proxy, "dummy", 123)

    with pytest.raises(RuntimeError, match="not fully initialized"):
        _ = uninitialized_dev_proxy["dummy"]

    with pytest.raises(RuntimeError, match="not fully initialized"):
        uninitialized_dev_proxy["dummy"] = 123

    with pytest.raises(RuntimeError, match="not fully initialized"):
        _ = "dummy" in uninitialized_dev_proxy

    with pytest.raises(RuntimeError, match="not fully initialized"):
        dir(uninitialized_dev_proxy)


def test_getattr_does_not_raise_for_removed_attribute():
    class TestDevice(Device):

        def initialize_dynamic_attributes(self):
            attr = attribute(
                name="dyn_attr",
                access=AttrWriteType.READ,
                fget=self.read_dyn_attr,
            )
            self.add_attribute(attr)

        def read_dyn_attr(self, attr) -> int:
            return 1

        @command
        def delete_attr(self):
            self.remove_attribute("dyn_attr")

    with DeviceTestContext(TestDevice) as proxy:
        assert hasattr(proxy, "dyn_attr")
        proxy.delete_attr()
        assert not hasattr(proxy, "dyn_attr")
